/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group (THG).                                         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of eos52nc4. The full eos52nc4 copyright notice,      *
 * including terms governing use, modification, and redistribution, is       *
 * contained in the files COPYING and Copyright.html. COPYING can be         *
 * found at the root of the source code distribution tree; Copyright.html    *
 * can be found at the root level of an installed copy of the electronic     *
 * eos52nc4 document set and is linked from the top-level documents page.   *
 * It can also be found at                                                   *
 * http://www.hdfgroup.org/H4H5TOOLS/doc/Copyright.html. If you do not have  *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define H5_USE_16_API 1
#include "HE5_HdfEosDef.h"
#include "toy_util.h"
#include "toy_hdfeos5.h"

#ifndef H5_USE_16_API
#error this file should be compiled with H5_USE_16_API
#endif


/** 
 * @brief call the passed @inq function and store the result in @dest
 * 
 * @param ctxid context identifier
 * @param fname HDF-EOS2 filename
 * @param inq inquiry API such as GDinqgrid, SWinqswath or PTinqpoint
 * @param[out] dest this function will allocate storage and fill the result of @inq
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_namelist(hid_t ctxid, const char *fname, long (*inq)(const char *, char *, long *), char **dest)
{
  long bufsize;
  int succeeded = 0;

  *dest = NULL;
  do {
    long numentries;

    if ((numentries = inq(fname, NULL, &bufsize)) == -1) err1_break(3, "inq fails")
    if (numentries == 0) bufsize = 0;
    if ((*dest = malloc(bufsize + 1)) == NULL) nomem_break()
    (*dest)[bufsize] = 0;
    if (numentries > 0) {
      if (inq(fname, *dest, &bufsize) == -1) err1_break(3, "inq_fails (2)")
    }

    succeeded = 1;
  } while (0);

  if (!succeeded) {
    if (*dest != NULL) free(*dest);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get EOS2 the list of dimensions by using the given set of functions
 * 
 * @param ctxid context identifier
 * @param id EOS2 id returned by GDattach or SWattach
 * @param entries GDnentries or SWnentries
 * @param inq GDinqdims or SWinqdims
 * @param[out] dims vector of dimension_t (pairs of name and size)
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_dimension(hid_t ctxid, hid_t id, long (*entries)(hid_t, int, long *), long (*inq)(hid_t, char *, hsize_t *), dimensionalloc *dims)
{
  char *namelist = NULL;
  hsize_t *dimsize = NULL;
  int succeeded = 0;

  do {
    long numdims, bufsize;

    if ((numdims = entries(id, HE5_HDFE_NENTDIM, &bufsize)) == -1) err1_break(3, "dimension entries failure")
    if (numdims > 0) {
      int i, j, failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("dimension namelist")
      namelist[bufsize] = 0;
      if ((dimsize = malloc(numdims * sizeof(hsize_t))) == NULL) nomem1_break("dimension size")
      if (inq(id, namelist, dimsize) == -1) err1_break(3, "dimension inq failure")

      for (i = j = 0; j <= bufsize; ++j) {
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  dimension_t *dim;
	  if (!dimensionalloc_readyplus(dims, 1)) nomem1_flagbreak(failinside, "dimension_t")
	  dim = &dims->s[dims->len++];
	  init_dimension_t(dim);
	  ASSERT(dims->len <= (size_t)numdims)
	  {
	    if (!challoc_copyb(&dim->name, namelist + i, j - i)) nomem1_flagbreak(failinside, "namelist")
	    dim->dimsize = dimsize[dims->len - 1];
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (namelist != NULL) free(namelist);
    if (dimsize != NULL) free(dimsize);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get field info of EOS2 grid or swath by using given set of functions
 * 
 * @param ctxid context identifier
 * @param id EOS2 id returned by GDattach or SWattach
 * @param entries GDnentries or SWnentries
 * @param inq GDinqfields, SWinqgeofields or SWinqdatafields
 * @param fldinfo GDfieldinfo or SWfieldinfo
 * @param inqtype HE5_GDinqdatatype
 * @param readfld GDreadfield or SWreadfield
 * @param getfill GDgetfillvalue or SWgetfillvalue
 * @param geofield determine the second argument of @entries function
 * If geofield is nonzero, HDFE_NENTGFLD is used for the second argument of @entries function; otherwise, HDFE_NENTFLD is used
 * @param getattr function to be called to fetch attributes e.g. get_gridlocalattribute
 * @param[out] fields vector of fieldinfo_t
 * 
 *
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_fieldinfo(hid_t ctxid, hid_t id, long (*entries)(hid_t, int, long *), long (*inq)(hid_t, char *, int *, hid_t *), herr_t (*fldinfo)(hid_t, const char *, int *, hsize_t *, hid_t *, char *, char *), herr_t (*inqtype)(hid_t, const char *, const char *, int, hid_t *, H5T_class_t *, H5T_order_t *, size_t *), herr_t (*readfld)(hid_t, const char *, const hssize_t *, const hsize_t *, const hsize_t *, void *), herr_t (*getfill)(hid_t, const char *, void *), int geofield, int (*getattr)(hid_t, hid_t, const char *, attributealloc *), fieldinfoalloc *fields)
{
  char *namelist = NULL;
  namelistalloc dimnames, maxdimnames;
  int succeeded = 0;

  namelistalloc_init(&dimnames);
  namelistalloc_init(&maxdimnames);
  do {
    long numfields, bufsize;

    if ((numfields = entries(id, geofield ? HE5_HDFE_NENTGFLD : HE5_HDFE_NENTDFLD, &bufsize)) == -1) err1_break(3, "fieldinfo entries failure")
    if (numfields > 0) {
      int i, j, failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("fieldinfo namelist")
      namelist[bufsize] = 0;
      if (inq(id, namelist, NULL, NULL) == -1) err1_break(3, "fieldinfo inq failure")

      for (i = j = 0; j <= bufsize; ++j) {
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  fieldinfo_t *field;
	  if (!fieldinfoalloc_readyplus(fields, 1)) nomem1_flagbreak(failinside, "fieldinfo_t")
	  field = &fields->s[fields->len++];
	  init_fieldinfo_t(field);
	  ASSERT(fields->len <= (size_t)numfields)
	  {
	    size_t k;
	    int failinside2 = 0;
	    hsize_t numelem = 0;
	    hsize_t dimsize[16]; // XXX: 16?
	    char dimlist[512], maxdimlist[512]; // This number came from one HDF-EOS developer
	    size_t typesize, elemsize;

	    if (!challoc_copyb(&field->name, namelist + i, j - i)) nomem1_flagbreak(failinside, "namelist")
	    /* The prototype of HE5_GDfieldinfo states that 5th argument is of hid_t[] type,
	     * but I think it should not be an array. Its code seems to write only the first
	     * element, and there is no way to recognize how many elements are available.
	     * I will guess only the first element will be used.
	     */
	    {
	      hid_t he5type;
	      if ((fldinfo(id, field->name.s, &field->rank, dimsize, &he5type, dimlist, maxdimlist)) == -1) err1_flagbreak(3, failinside, "fieldinfo")
	    }

	    namelistalloc_empty(&dimnames);
	    namelistalloc_empty(&maxdimnames);
	    if (parse_list(dimlist, strlen(dimlist), ',', &dimnames) == FAIL) err_flagbreak(3, failinside)
	    if (parse_list(maxdimlist, strlen(maxdimlist), ',', &maxdimnames) == FAIL) err_flagbreak(3, failinside)
	    /* assert that rank is equal to number of names from comma-separated dimension list */
	    if (dimnames.len != field->rank) err1_flagbreak(3, failinside, "rank != dimension name list")
	    if (maxdimnames.len != field->rank) err1_flagbreak(3, failinside, "rank != max dimension name list")
	    for (k = 0; k < dimnames.len; ++k) {
	      dimension_t *dim;
	      if (!dimensionalloc_readyplus(&field->dims, 1)) nomem1_flagbreak(failinside2, "dimension_t extension")
	      dim = &field->dims.s[field->dims.len++];
	      init_dimension_t(dim);
	      {
		if (!challoc_copy(&dim->name, &dimnames.s[k].name)) nomem1_flagbreak(failinside2, "dimension name copy")
		if (!challoc_copy(&dim->max, &maxdimnames.s[k].name)) nomem1_flagbreak(failinside2, "max dimension name copy")
		dim->dimsize = dimsize[k];
	      }
	    }
	    if (failinside2) { failinside = failinside2; break; }

	    for (k = 0, numelem = 1; k < field->dims.len; ++k)
	      numelem *= field->dims.s[k].dimsize;

	    {
	      H5T_class_t classid;
	      H5T_order_t order;
	      if (inqtype(id, field->name.s, NULL, geofield ? HE5_HDFE_GEOGROUP : HE5_HDFE_DATAGROUP, &field->type, &classid, &order, &typesize) == -1) err1_flagbreak(3, failinside, "inqtype")
	    }

	    if ((elemsize = H5Tget_size(field->type)) == 0) err1_flagbreak(3, failinside, "H5Tget_size")

#ifdef EAGER_DATA_FETCH
	    /* read data */
	    if (!challoc_ready(&field->data, numelem * elemsize)) nomem1_flagbreak(failinside, "fieldinfo data")
	    if (readfld(id, field->name.s, NULL, NULL, NULL, field->data.s) == -1) err1_flagbreak(3, failinside, "readfield failure")
	    field->data.len = (size_t)numelem * elemsize;
#else
	    field->datalen = (size_t)numelem * elemsize;
#endif

	    /* filler may not exist */
	    if (!challoc_ready(&field->filler, elemsize)) nomem1_flagbreak(failinside, "fieldinfo filler")
	    if (getfill(id, field->name.s, field->filler.s) != -1)
	      field->filler.len = elemsize;

	    /* attribute */
	    if (getattr) {
	      if (getattr(ctxid, id, field->name.s, &field->localattr) == -1) err_flagbreak(3, failinside)
	    }
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (namelist != NULL) free(namelist);
    namelistalloc_free(&dimnames);
    namelistalloc_free(&maxdimnames);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get a vector of attributes of EOS5 global, grid or swath by using given set of functions
 * TODO: I think nobody likes this function which has too many parameters. Is there any type-safe and generic way?
 * 
 * @param ctxid context identifier
 * @param id EOS2 id returned by GDattach or SWattach
 * @param kind global or HE5_HDFE_ATTRGROUP, HE5_HDFE_GRPATTRGROUP, HE5_HDFE_LOCATTRGROUP
 * @param fieldname field name for local attributes
 * @param inq3      HE5_EHinqglbattrs, HE5_GDinqattrs, HE5_GDinqgrpattrs,
 * @param inq4                                                            HE5_GDinqlocattrs
 * @param attrinfo4 HE5_EHglbattrinfo, HE5_GDattrinfo, HE5_GDgrpattrinfo,
 * @param attrinfo5                                                       HE5_GDlocattrinfo
 * @param inqtype6 HE5_EHinqglbdatatype,
 * @param inqtype8                       HE5_GDinqdatatype
 * @param readattr3 HE5_EHreadglbattr, HE5_GDreadattr, HE5_GDreadgrpattr
 * @param readattr4                                                       HE5_GDreadlocattr
 * @param[out] attrs vector of attributes
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_attribute(hid_t ctxid, hid_t id, int kind, const char *fieldname,
  long (*inq3)(hid_t, char *, long *),
  long (*inq4)(hid_t, const char *, char *, long *),
  herr_t (*attrinfo4)(hid_t, const char *, hid_t *, hsize_t *),
  herr_t (*attrinfo5)(hid_t, const char *, const char *, hid_t *, hsize_t *),
  herr_t (*inqtype6)(hid_t, const char *, hid_t *, H5T_class_t *, H5T_order_t *, size_t *),
  herr_t (*inqtype8)(hid_t, const char *, const char *, int, hid_t *, H5T_class_t *, H5T_order_t *, size_t *),
  herr_t (*readattr3)(hid_t, const char *, void *),
  herr_t (*readattr4)(hid_t, const char *, const char *, void *),
  attributealloc *attrs)
{
  char *namelist = NULL;
  int succeeded = 0;

  ASSERT((inq3 == 0) ^ (inq4 == 0))
  ASSERT((attrinfo4 == 0) ^ (attrinfo5 == 0))
  ASSERT((inqtype6 == 0) ^ (inqtype8 == 0))
  ASSERT((readattr3 == 0) ^ (readattr4 == 0))

  do {
    long numattrs, bufsize;

    if (inq3) {
      if ((numattrs = inq3(id, NULL, &bufsize)) == -1) err1_break(ctxid, "attribute inq failure")
    }
    else {
      if ((numattrs = inq4(id, fieldname, NULL, &bufsize)) == -1) err1_break(ctxid, "attribute inq failure")
    }
    if (numattrs > 0) {
      int i, j, failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("attribute namelist")
      namelist[bufsize] = 0;
      if (inq3) {
	if (inq3(id, namelist, &bufsize) == -1) err1_break(ctxid, "attribute inq failure (2)")
      }
      else {
	if (inq4(id, fieldname, namelist, &bufsize) == -1) err1_break(ctxid, "attribute inq failure (2)")
      }

      for (i = j = 0; j <= bufsize; ++j) {
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  attribute_t *attr;
	  if (!attributealloc_readyplus(attrs, 1)) nomem1_flagbreak(failinside, "attribute_t")
	  attr = &attrs->s[attrs->len++];
	  init_attribute_t(attr);
	  ASSERT(attrs->len <= (size_t)numattrs)
	  {
	    hsize_t count;
	    size_t typesize;

	    if (!challoc_copyb(&attr->name, namelist + i, j - i)) nomem1_flagbreak(failinside, "attribute name")
	    {
	      hid_t he5type;
	      if (attrinfo4) {
		if (attrinfo4(id, attr->name.s, &he5type, &count) == -1) err1_flagbreak(ctxid, failinside, "attrinfo failure")
	      }
	      else {
		if (attrinfo5(id, fieldname, attr->name.s, &he5type, &count) == -1) err1_flagbreak(ctxid, failinside, "attrinfo failure")
	      }
	    }
	    {
	      H5T_class_t classid;
	      H5T_order_t order;
	      if (inqtype6) {
		if (inqtype6(id, attr->name.s, &attr->type, &classid, &order, &typesize) == -1) err1_flagbreak(3, failinside, "inqtype")
	      }
	      else {
		if (inqtype8(id, fieldname, attr->name.s, kind, &attr->type, &classid, &order, &typesize) == -1) err1_flagbreak(3, failinside, "inqtype")
	      }
	    }

	    if (!challoc_ready(&attr->value, (size_t)count * typesize)) nomem1_flagbreak(failinside, "attribute value")
	    if (readattr3) {
	      if (readattr3(id, attr->name.s, attr->value.s) == -1) err1_flagbreak(ctxid, failinside, "readattr failure")
	    }
	    else {
	      if (readattr4(id, fieldname, attr->name.s, attr->value.s) == -1) err1_flagbreak(ctxid, failinside, "readattr failure")
	    }
	    attr->value.len = (size_t)count * typesize;
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (namelist != NULL) free(namelist);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get EOS2 grid information by using GDgridinfo API
 * 
 * @param ctxid context identifier
 * @param gdid EOS2 id returned by GDattach
 * @param[out] info result
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_gridinfo(hid_t ctxid, hid_t gdid, gridinfo_t *info)
{
  if (HE5_GDgridinfo(gdid, &info->xdim, &info->ydim, info->upleft, info->lowright) == -1)
    err1_return(3, "GDgridinfo failure")
  return SUCCEED;
}

/** 
 * @brief get projection information by using GDprojinfo API
 * 
 * @param ctxid context identifier
 * @param gdid EOS2 id returned by GDattach
 * @param grid grid information from GDgridinfo API
 * @param[out] proj projection information
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_gridprojinfo(hid_t ctxid, hid_t gdid, const gridinfo_t *grid, projinfo_t *proj)
{
  if (HE5_GDprojinfo(gdid, &proj->proj, &proj->zone, &proj->sphere, proj->param) == -1)
    err1_return(3, "GDprojiunfo failure")
  if (HE5_GDpixreginfo(gdid, &proj->pix) == -1) {
    proj->pix = HE5_HDFE_CENTER;
    warn1(3, "GDpixreginfo failure: assuming HE5_HDFE_CENTER")
  }
  if (HE5_GDorigininfo(gdid, &proj->origin) == -1) {
    proj->origin = HE5_HDFE_GD_UL;
    warn1(3, "GDorigininfo failure: assuming HE5_HDFE_GD_UL")
  }

  return SUCCEED;
}

static int get_swathdimmap(hid_t ctxid, hid_t swid, dimmapalloc *maps)
{
  char *namelist = NULL;
  long *offset = NULL;
  long *increment = NULL;
  namelistalloc name;
  int succeeded = 0;

  namelistalloc_init(&name);
  do {
    long nummaps, bufsize;

    if ((nummaps = HE5_SWnentries(swid, HE5_HDFE_NENTMAP, &bufsize)) == -1) err1_break(3, "dimmap entries")
    if (nummaps > 0) {
      int i, j, failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("dimmap namelist")
      if ((offset = malloc(nummaps * sizeof(long))) == NULL) nomem1_break("dimmap offset")
      if ((increment = malloc(nummaps * sizeof(long))) == NULL) nomem1_break("dimmap increment")
      if (HE5_SWinqmaps(swid, namelist, offset, increment) == -1) err1_break(3, "dimmap")

      for (i = j = 0; j <= bufsize; ++j) {
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  dimmap_t *map;
	  if (!dimmapalloc_readyplus(maps, 1)) nomem1_flagbreak(failinside, "dimmap")
	  map = &maps->s[maps->len++];
	  init_dimmap_t(map);
	  ASSERT(maps->len <= (size_t)nummaps)
	  {
	    namelistalloc_empty(&name);
	    if (parse_list(namelist + i, j - i, '/', &name) == FAIL) err_flagbreak(3, failinside)
	    ASSERT(name.len == 2)
	    if (!challoc_copy(&map->geo, &name.s[0].name)) nomem1_flagbreak(failinside, "indexmap geo")
	    if (!challoc_copy(&map->data, &name.s[1].name)) nomem1_flagbreak(failinside, "indexmap data")
	    map->offset = offset[maps->len - 1];
	    map->increment = increment[maps->len - 1];
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    namelistalloc_free(&name);
    if (namelist) free(namelist);
    if (offset) free(offset);
    if (increment) free(increment);
  }
  return succeeded ? SUCCEED : FAIL;
}

static int get_swathindexmap(hid_t ctxid, hid_t swid, indexmapalloc *indices)
{
  char *namelist = NULL;
  namelistalloc name;
  int succeeded = 0;

  namelistalloc_init(&name);
  do {
    long numindices, bufsize;

    if ((numindices = HE5_SWnentries(swid, HE5_HDFE_NENTIMAP, &bufsize)) == -1) err1_break(3, "SWnentries HDFE_NENTIMAP failure")
    if (numindices > 0) {
      /* TODO: I have never seen any EOS5 files that have index map. */
      int i, j, failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("indexmap namelist")
      if (HE5_SWinqidxmaps(swid, namelist, NULL) == -1) err1_break(3, "SWinqdimmaps failure")

      for (i = j = 0; j <= bufsize; ++j) {
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  indexmap_t *index;
	  if (!indexmapalloc_readyplus(indices, 1)) nomem1_flagbreak(failinside, "indexmap_t")
	  index = &indices->s[indices->len++];
	  init_indexmap_t(index);
	  ASSERT(indices->len <= (size_t)numindices)
	  {
	    hsize_t dimsize;
	    namelistalloc_empty(&name);
	    if (parse_list(namelist + i, j - i, '/', &name) == FAIL) err_flagbreak(3, failinside)
	    ASSERT(name.len == 2)
	    if (!challoc_copy(&index->geo, &name.s[0].name)) nomem1_flagbreak(failinside, "indexmap geo")
	    if (!challoc_copy(&index->data, &name.s[1].name)) nomem1_flagbreak(failinside, "indexmap data")

	    if ((dimsize = HE5_SWdiminfo(swid, index->geo.s)) == -1) err1_flagbreak(3, failinside, "SWdiminfo failure")
	    if (!longalloc_ready(&index->indices, (size_t)dimsize)) nomem1_flagbreak(failinside, "indexmap data")
	    if (HE5_SWidxmapinfo(swid, index->geo.s, index->data.s, index->indices.s) == -1) err1_flagbreak(3, failinside, "SWidxmapinfo failure")
	    index->indices.len = (size_t)dimsize;
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    namelistalloc_free(&name);
    if (namelist != NULL) free(namelist);
  }
  return succeeded ? SUCCEED : FAIL;
}

static long HE5_GDinqdims_long(hid_t gdid, char *dimnames, hsize_t dims[])
{
  return HE5_GDinqdims(gdid, dimnames, dims);
}

static int get_griddimension(hid_t ctxid, hid_t gdid, dimensionalloc *dims)
{
  return get_dimension(ctxid, gdid, HE5_GDnentries, HE5_GDinqdims_long, dims);
}

static int get_globalattribute(hid_t ctxid, hid_t eosfd, attributealloc *attrs)
{
  return get_attribute(ctxid, eosfd, 0, NULL, HE5_EHinqglbattrs, NULL, HE5_EHglbattrinfo, NULL, HE5_EHinqglbdatatype, NULL, HE5_EHreadglbattr, NULL, attrs);
}

static int get_gridattribute(hid_t ctxid, hid_t gdid, attributealloc *attrs)
{
  return get_attribute(ctxid, gdid, HE5_HDFE_ATTRGROUP, NULL, HE5_GDinqattrs, NULL, HE5_GDattrinfo, NULL, NULL, HE5_GDinqdatatype, HE5_GDreadattr, NULL, attrs);
}

static int get_gridgroupattribute(hid_t ctxid, hid_t gdid, attributealloc *attrs)
{
  return get_attribute(ctxid, gdid, HE5_HDFE_GRPATTRGROUP, NULL, HE5_GDinqgrpattrs, NULL, HE5_GDgrpattrinfo, NULL, NULL, HE5_GDinqdatatype, HE5_GDreadgrpattr, NULL, attrs);
}

static int get_gridlocalattribute(hid_t ctxid, hid_t gdid, const char *fieldname, attributealloc *attrs)
{
  return get_attribute(ctxid, gdid, HE5_HDFE_LOCATTRGROUP, fieldname, NULL, HE5_GDinqlocattrs, NULL, HE5_GDlocattrinfo, NULL, HE5_GDinqdatatype, NULL, HE5_GDreadlocattr, attrs);
}

static long HE5_GDinqfields_long(hid_t gdid, char *fieldlist, int rank[], hid_t ntype[])
{
  return HE5_GDinqfields(gdid, fieldlist, rank, ntype);
}

static int get_gridfieldinfo(hid_t ctxid, hid_t gdid, fieldinfoalloc *fields)
{
  return get_fieldinfo(ctxid, gdid, HE5_GDnentries, HE5_GDinqfields_long, HE5_GDfieldinfo, HE5_GDinqdatatype, HE5_GDreadfield, HE5_GDgetfillvalue, 0, get_gridlocalattribute, fields);
}

static int get_swathdimension(hid_t ctxid, hid_t swid, dimensionalloc *dims)
{
  return get_dimension(ctxid, swid, HE5_SWnentries, HE5_SWinqdims, dims);
}

static int get_swathattribute(hid_t ctxid, hid_t swid, attributealloc *attrs)
{
  return get_attribute(ctxid, swid, HE5_HDFE_ATTRGROUP, NULL, HE5_SWinqattrs, NULL, HE5_SWattrinfo, NULL, NULL, HE5_SWinqdatatype, HE5_SWreadattr, NULL, attrs);
}

static int get_swathgroupattribute(hid_t ctxid, hid_t swid, attributealloc *attrs)
{
  return get_attribute(ctxid, swid, HE5_HDFE_GRPATTRGROUP, NULL, HE5_SWinqgrpattrs, NULL, HE5_SWgrpattrinfo, NULL, NULL, HE5_SWinqdatatype, HE5_SWreadgrpattr, NULL, attrs);
}

static int get_swathlocalattribute(hid_t ctxid, hid_t swid, const char *fieldname, attributealloc *attrs)
{
  return get_attribute(ctxid, swid, HE5_HDFE_LOCATTRGROUP, fieldname, NULL, HE5_SWinqlocattrs, NULL, HE5_SWlocattrinfo, NULL, HE5_SWinqdatatype, NULL, HE5_SWreadlocattr, attrs);
}

static herr_t HE5_SWfieldinfo_const(hid_t swid, const char *fieldname, int *rank, hsize_t dims[], hid_t ntype[], char *dimlist, char *maxdimlist)
{
  return HE5_SWfieldinfo(swid, (char *)fieldname, rank, dims, ntype, dimlist, maxdimlist);
}

static herr_t HE5_SWreadfield_const(hid_t swid, const char *fieldname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[],  void *data)
{
  return HE5_SWreadfield(swid, (char *)fieldname, start, stride, edge, data);
}

static herr_t HE5_SWgetfillvalue_const(hid_t swid, const char *fieldname, void *fillval)
{
  return HE5_SWgetfillvalue(swid, (char *)fieldname, fillval);
}

static int get_swathfieldinfo(hid_t ctxid, hid_t swid, int geofield, fieldinfoalloc *fields)
{
  return get_fieldinfo(ctxid, swid, HE5_SWnentries, geofield ? HE5_SWinqgeofields : HE5_SWinqdatafields, HE5_SWfieldinfo_const, HE5_SWinqdatatype, HE5_SWreadfield_const, HE5_SWgetfillvalue_const, geofield, get_swathlocalattribute, fields);
}

/** 
 * @brief opens the given HDF-EOS5 file and fetches the list of grids and swaths
 * 
 * @param ctxid context identifier
 * @param fname the HDF-EOS5 filename to be opened
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
int open_hdfeos5(hid_t ctxid, const char *fname)
{
  hid_t gdfd = -1;
  hid_t swfd = -1;
  char *gridlist = NULL;
  char *swathlist = NULL;
  char *attrlist = NULL;
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;

  ASSERT(ctx->gridfd == -1)
  ASSERT(ctx->swathfd == -1)
  ASSERT(ctx->gridid == -1)
  ASSERT(ctx->swathid == -1)
  do {
    if (ctx == NULL) err1_break(3, "invalid context id")

    if ((gdfd = HE5_GDopen(fname, H5F_ACC_RDONLY)) == FAIL) err1_break(3, "HE5_GDopen")
    if (get_namelist(ctxid, fname, HE5_GDinqgrid, &gridlist) == FAIL) err1_break(3, "HE5_GDinqgrid")
    if (parse_list(gridlist, strlen(gridlist), ',', &ctx->gridlist) == FAIL) err_break(3)

    if ((swfd = HE5_SWopen(fname, H5F_ACC_RDONLY)) == FAIL) err1_break(3, "HE5_SWopen")
    if (get_namelist(ctxid, fname, HE5_SWinqswath, &swathlist) == FAIL) err1_break(3, "HE5_SWinqswath")
    if (parse_list(swathlist, strlen(swathlist), ',', &ctx->swathlist) == FAIL) err_break(3)

    if (get_globalattribute(ctxid, gdfd, &ctx->globalattr) == FAIL) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    if (gridlist != NULL) free(gridlist);
    if (swathlist != NULL) free(swathlist);
    if (attrlist != NULL) free(attrlist);
  }
  if (succeeded) {
    ctx->infname = fname;
    ctx->gridfd = gdfd;
    ctx->swathfd = swfd;
  }
  else {
    if (gdfd != -1) HE5_GDclose(gdfd);
    if (swfd != -1) HE5_SWclose(swfd);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief read data in HDF-EOS5 grid
 * 
 * @param ctxid context identifier
 * @param gridname grid name
 * @param[out] grid all grid information retrieved from the HDF-EOS5 library
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
int read_hdfeos5_onegrid(hid_t ctxid, const char *gridname, grid_t *grid)
{
  hid_t gdid = -1;
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;

  init_grid_t(grid);
  do {
    if (ctx == NULL) err1_break(3, "invalid context id")

    if (ctx->gridid != -1) {
      HE5_GDdetach(ctx->gridid);
      ctx->gridid = -1;
    }

    if ((gdid = HE5_GDattach(ctx->gridfd, gridname)) == FAIL) err1_break(3, "GDattach failure")
    if (get_gridinfo(ctxid, gdid, &grid->info) == FAIL) err_break(3)
    if (get_gridprojinfo(ctxid, gdid, &grid->info, &grid->proj) == FAIL) err_break(3)
    if (get_griddimension(ctxid, gdid, &grid->dim) == FAIL) err_break(3)
    if (get_gridfieldinfo(ctxid, gdid, &grid->field) == FAIL) err_break(3)
    if (get_gridattribute(ctxid, gdid, &grid->attr) == FAIL) err_break(3)
    if (get_gridgroupattribute(ctxid, gdid, &grid->grpattr) == FAIL) err_break(3)

    succeeded = 1;
  } while (0);

  if (succeeded) {
    ctx->gridid = gdid;
  }
  else {
    free_grid_t(grid);
    if (gdid != -1) HE5_GDdetach(gdid);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief read data in HDF-EOS2 swath
 * 
 * @param ctxid context identifier
 * @param swathname swath name
 * @param[out] swath all swath information retrieved from the HDF-EOS5 library
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
int read_hdfeos5_oneswath(hid_t ctxid, const char *swathname, swath_t *swath)
{
  hid_t swid = -1;
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;

  init_swath_t(swath);
  do {
    if (ctx == NULL) err1_break(3, "invalid context id")

    if (ctx->swathid != -1) {
      HE5_SWdetach(ctx->swathid);
      ctx->swathid = -1;
    }

    if ((swid = HE5_SWattach(ctx->swathfd, swathname)) == FAIL) err1_break(3, "SWattach failure")
    if (get_swathdimension(ctxid, swid, &swath->dim) == -1) err_break(3)
    if (get_swathdimmap(ctxid, swid, &swath->dimmap) == -1) err_break(3)
    if (get_swathindexmap(ctxid, swid, &swath->index) == -1) err_break(3)
    if (get_swathfieldinfo(ctxid, swid, 1, &swath->geofield) == -1) err_break(3)
    if (get_swathfieldinfo(ctxid, swid, 0, &swath->datafield) == -1) err_break(3)
    if (get_swathattribute(ctxid, swid, &swath->attr) == -1) err_break(3)
    if (get_swathgroupattribute(ctxid, swid, &swath->grpattr) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (succeeded) {
    ctx->swathid = swid;
  }
  else {
    free_swath_t(swath);
    if (swid != -1) HE5_SWdetach(swid);
  }
  return succeeded ? SUCCEED : FAIL;
}

#ifdef EAGER_DATA_FETCH
#error This program requires lazy data fetch.
#else
int prepare_field_lazydata(hid_t ctxid, int isgrid, int isgeo, fieldinfo_t *field)
{
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);

    if (field->lazydata) {
      ASSERT(field->lazydata->len == field->datalen);
    }
    else {
      herr_t r = -1;
      if ((field->lazydata = malloc(sizeof(challoc))) == NULL) nomem1_break("lazydata")
      challoc_init(field->lazydata);
      if (!challoc_ready(field->lazydata, field->datalen)) nomem1_break("lazydata elem")

      if (isgrid) {
	ASSERT(!isgeo)
	r = HE5_GDreadfield(ctx->gridid, field->name.s, NULL, NULL, NULL, field->lazydata->s);
      }
      else {
	r = HE5_SWreadfield_const(ctx->swathid, field->name.s, NULL, NULL, NULL, field->lazydata->s);
      }
      if (r == -1) err2_break(3, "readfield error", field->name.s)
      field->lazydata->len = field->datalen;
    }

    succeeded = 1;
  } while (0);
  return succeeded ? 0 : -1;
}

int drop_field_lazydata(hid_t ctxid, fieldinfo_t *field)
{
  if (field->lazydata) {
    challoc_free(field->lazydata);
    free(field->lazydata);
    field->lazydata = NULL;
  }
  return 0;
}
#endif

int close_hdfeos5(hid_t ctxid)
{
  context_t *ctx = get_context(ctxid);
  if (ctx != NULL) {
    if (ctx->gridid != -1) HE5_GDdetach(ctx->gridid);
    ctx->gridid = -1;
    if (ctx->swathid != -1) HE5_SWdetach(ctx->swathid);
    ctx->swathid = -1;

    if (ctx->gridfd != -1) HE5_GDclose(ctx->gridfd);
    ctx->gridfd = -1;
    if (ctx->swathfd != -1) HE5_SWclose(ctx->swathfd);
    ctx->swathfd = -1;
    return 0;
  }
  return -1;
}

// vim:ts=8:sw=2:sts=2:cindent
