/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group (THG).                                         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of eos52nc4. The full eos52nc4 copyright notice,      *
 * including terms governing use, modification, and redistribution, is       *
 * contained in the files COPYING and Copyright.html. COPYING can be         *
 * found at the root of the source code distribution tree; Copyright.html    *
 * can be found at the root level of an installed copy of the electronic     *
 * eos52nc4 document set and is linked from the top-level documents page.   *
 * It can also be found at                                                   *
 * http://www.hdfgroup.org/H4H5TOOLS/doc/Copyright.html. If you do not have  *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "toy_util.h"
#include "toy_hdf5.h"
#include "H5Gpublic.h"
#include "H5Ppublic.h"
#include "H5Spublic.h"
#include "H5DSpublic.h"
#include "H5Apublic.h"

#ifdef H5_USE_16_API
#error this file should be compiled without H5_USE_16_API
#endif

/* this should be removed in the release */
#ifdef _WIN32
#define MAKE_HDFVIEW_HAPPY
#endif

#define _C_VECTOR_IMPL
#include "toy_vector.h"

static const char *DATAGROUP_NAME = "Data Fields";
static const char *GEOGROUP_NAME = "Geolocation Fields";

static hid_t get_h5group(hid_t ctxid, const char *path)
{
  hid_t groupid = -1;
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);
    if ((groupid = H5Gopen(ctx->outfd, path, H5P_DEFAULT)) == -1) err2_break(3, "cannot open h5group", path)

    succeeded = 1;
  } while (0);

  if (!succeeded) {
    if (groupid != -1) H5Gclose(groupid);
  }
  return succeeded ? groupid : -1;
}

static int nc4safename(hid_t ctxid, challoc *name)
{
  challoc buffer;
  int succeeded = 0;

  challoc_init(&buffer);
  do {
    int failinside = 0;
    size_t i;

#define IS_DIGIT(c)  ('0' <= (c) && (c) <= '9')
#define IS_ALPHA(c)  (('a' <= (c) && (c) <= 'z') || ('A' <= (c) && (c) <= 'Z'))
#define IS_OTHER(c)  ((c) == '_')

    for (i = 0; i < name->len; ++i) {
      if (buffer.len == 0) {
	if (IS_DIGIT(name->s[i])) {
	  if (!challoc_append(&buffer, "_")) nomem1_flagbreak(failinside, "prepend _")
	}
      }

      if (IS_DIGIT(name->s[i]) || IS_ALPHA(name->s[i]) || IS_OTHER(name->s[i])) {
	if (!challoc_append(&buffer, &name->s[i])) nomem1_flagbreak(failinside, "a character")
      }
      else {
	if (!challoc_append(&buffer, "_")) nomem1_flagbreak(failinside, "copy _")
      }
    }
    if (failinside) break;

    if (!challoc_copy(name, &buffer)) nomem1_break("buffer")

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&buffer);
  }
  return succeeded ? 0 : -1;
}

static int nc4safenamecopy(hid_t ctxid, const char *eosname, challoc *nc4name)
{
  int succeeded = 0;

  do {
    if (!challoc_copys(nc4name, eosname)) nomem1_break("safename")
    if (nc4safename(ctxid, nc4name) == -1) err_break(3)

    succeeded = 1;
  } while (0);
  return succeeded ? 0 : -1;
}

static int nc4safeattrname(hid_t ctxid, challoc *name)
{
  return 0;
}

static int nc4safeattrnamecopy(hid_t ctxid, const char *eosname, challoc *nc4name)
{
  int succeeded = 0;

  do {
    if (!challoc_copys(nc4name, eosname)) nomem1_break("safe attr name")
    if (nc4safeattrname(ctxid, nc4name) == -1) err_break(3)

    succeeded = 1;
  } while (0);
  return succeeded ? 0 : -1;
}

static hid_t creat_h5group(hid_t ctxid, const char *parentpath, const challoc *nodename)
{
  hid_t parentid = -1;
  hid_t propid = -1;
  hid_t thisid = -1;
  int succeeded = 0;

  do {
#ifdef MAKE_HDFVIEW_HAPPY
    if ((parentid = get_h5group(ctxid, parentpath)) == -1) err3_break(3, "cannot create h5group", parentpath, nodename->s)
    if ((thisid = H5Gcreate(parentid, nodename->s, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0) err1_break(3, "H5Gcreate_anon")
#else
    if ((parentid = get_h5group(ctxid, parentpath)) == -1) err3_break(3, "cannot create h5group", parentpath, nodename->s)
    if ((propid = H5Pcreate(H5P_GROUP_CREATE)) == -1) err1_break(3, "H5Pcreate")
    if (H5Pset_link_creation_order(propid, H5P_CRT_ORDER_TRACKED | H5P_CRT_ORDER_INDEXED) < 0) err1_break(3, "H5Pset_link_creation_order")
    if (H5Pset_attr_creation_order(propid, H5P_CRT_ORDER_TRACKED | H5P_CRT_ORDER_INDEXED) < 0) err1_break(3, "H5Pset_attr_creation_order")
    if ((thisid = H5Gcreate_anon(parentid, propid, H5P_DEFAULT)) < 0) err1_break(3, "H5Gcreate_anon")
    if ((H5Olink(thisid, parentid, nodename->s, H5P_DEFAULT, H5P_DEFAULT)) < 0) err1_break(3, "H5Link")
#endif

    succeeded = 1;
  } while (0);

  if (1) {
    if (propid != -1) H5Pclose(propid);
    if (parentid != -1) H5Gclose(parentid);
  }
  if (!succeeded) {
    if (thisid != -1) H5Gclose(thisid);
  }
  return succeeded ? thisid : -1;
}

static int detect_stringtype(hid_t ctxid, hid_t type)
{
  H5T_class_t c = H5Tget_class(type);
  if (c == H5T_STRING)
    return 1;
  return 0;
}

static int write_attribute(hid_t ctxid, hid_t target, const char *eosname, hid_t type, const void *data, size_t datalen)
{
  hid_t space = -1, at = -1;
  challoc nc4attrname;
  int succeeded = 0;

  challoc_init(&nc4attrname);
  do {
    size_t typesize;

    if (nc4safeattrnamecopy(ctxid, eosname, &nc4attrname) == -1) err_break(3)
    if (detect_stringtype(ctxid, type)) {
      if ((space = H5Screate(H5S_SCALAR)) == -1) err1_break(3, "space create")
    }
    else {
      hsize_t count;
      if ((typesize = H5Tget_size(type)) == 0) err1_break(3, "unknown type size")
      count = datalen / typesize;
      if ((space = H5Screate_simple(1, &count, &count)) == -1) err1_break(3, "space create")
    }
    if ((at = H5Acreate(target, nc4attrname.s, type, space, H5P_DEFAULT, H5P_DEFAULT)) < 0) err1_break(3, "attribute create")
    if (H5Awrite(at, type, data) == -1) err1_break(3, "attribute write")

    succeeded = 1;
  } while (0);

  if (1) {
    if (at != -1) H5Aclose(at);
    if (space != -1) H5Sclose(space);
    challoc_free(&nc4attrname);
  }
  return succeeded ? 0 : -1;
}

static int write_attribute_string(hid_t ctxid, hid_t dsetid, const char *eosname, const char *value)
{
  hid_t tid = -1;
  int succeeded = 0;

  do {
    size_t valuelen = strlen(value);

    if ((tid = H5Tcopy(H5T_C_S1)) == -1) err1_break(3, "type create")
    if (H5Tset_size(tid, strlen(value)) == -1) err1_break(3, "type size")
    if (H5Tset_strpad(tid, H5T_STR_SPACEPAD) == -1) err1_break(3, "type pad")

    if (write_attribute(ctxid, dsetid, eosname, tid, value, valuelen) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    if (tid != -1) H5Tclose(tid);
  }
  return succeeded ? 0 : -1;
}

typedef struct {
  const char *name;
  const char *longname;
  const char *unit;
  const char *axis;
  int variable;
  int dimension;
} coordinate_t;

static const coordinate_t * get_fake_coordinate(const grid_t *grid, int x)
{
  static const coordinate_t lon = { "lon", "longitude", "degrees_east", NULL, 1, 0 };
  static const coordinate_t lat = { "lat", "latitude", "degrees_north", NULL, 1, 0 };

  const coordinate_t *ret = x ? &lon : &lat;
  ASSERT(ret)
  /* Section 4.1 of CF convention states the following:
   * Variables representing latitude (or longitude) must always explicitly include
   * the "units" attribute */
  ASSERT(ret->unit)
  return ret;
}

static const coordinate_t * get_coordinate(const grid_t *grid, int orthogonal, int x)
{
  static const coordinate_t lon = { "lon", "longitude", "degrees_east", NULL, 1, 1 };
  static const coordinate_t lat = { "lat", "latitude", "degrees_north", NULL, 1, 1 };
  static const coordinate_t xdim = { "XDim", "x-coordinate in Cartesian system", NULL, "X", 0, 1 };
  static const coordinate_t ydim = { "YDim", "y-coordinate in Cartesian system", NULL, "Y", 0, 1 };

  const coordinate_t *ret = NULL;
  /* If the calculated lon and lat are orthogonal, it means lon and lat are independent.
   * For this case, XDim and YDim are actual dimensional scales; otherwise, they are just
   * like link between variables and 2-dimensional lon and lat. */
  if (orthogonal)
    ret = x ? &lon : &lat;
  else
    ret = x ? &xdim : &ydim;
  ASSERT(ret)
  return ret;
}

static void get_coordinate_from_dimension(const char *dimname, coordinate_t *coord)
{
  /* write_dimscale() can be used for both coordinate variables and normal dimensions.
   * This function wraps dimension name with coordinate_t so that write_dimscale()
   * can accept this dimension. */
  coord->name = dimname;
  coord->longname = NULL;
  coord->unit = NULL;
  coord->axis = NULL;
  coord->variable = 0;
  coord->dimension = 1;
}

static void get_coordinate_from_variable(const char *varname, coordinate_t *coord)
{
  coord->name = varname;
  coord->longname = NULL;
  coord->unit = NULL;
  coord->axis = NULL;
  coord->variable = 1;
  coord->dimension = 0;
}

static int write_dimscale_attrs(hid_t ctxid, hid_t target, const coordinate_t *coord)
{
  int succeeded = 0;

  do {
    if (coord->longname) {
      if (write_attribute_string(ctxid, target, "long_name", coord->longname) == -1) err1_break(3, "long_name attr")
    }
    if (coord->unit) {
      if (write_attribute_string(ctxid, target, "units", coord->unit) == -1) err1_break(3, "units attr")
    }
    if (coord->axis) {
      if (write_attribute_string(ctxid, target, "axis", coord->axis) == -1) err1_break(3, "axis attr")
    }

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int create_dimscale2(hid_t ctxid, const challoc *gridgrouppath, const challoc *nc4dimname, const coordinate_t *coord, hsize_t dimlength)
{
  hid_t gridgroup = -1, dimscaleset = -1, space = -1, propid = -1;
  int succeeded = 0;

  ASSERT(coord->dimension)
  do {
    hid_t filetype = H5T_IEEE_F64BE;
    const char *nc4dimspecial = coord->variable ? NULL : "This is a netCDF dimension but not a netCDF variable.";
    hsize_t maxdimlength = dimlength;

    if (dimlength == H5S_UNLIMITED)
      dimlength = 0;

    if ((gridgroup = get_h5group(ctxid, gridgrouppath->s)) == -1) err_break(3)

    if ((space = H5Screate_simple(1, &dimlength, &maxdimlength)) == -1) err2_break(3, "space create", nc4dimname->s)
    if ((propid = H5Pcreate(H5P_DATASET_CREATE)) < 0) err1_break(3, "creation property")
    if (maxdimlength == H5S_UNLIMITED) {
      hsize_t chunksize = 1;
      if (H5Pset_chunk(propid, 1, &chunksize) < 0) err1_break(3, "H5Pset_chunk")
    }
    if ((dimscaleset = H5Dcreate(gridgroup, nc4dimname->s, filetype, space, H5P_DEFAULT, propid, H5P_DEFAULT)) == -1) err2_break(3, "dimscale dataset create", nc4dimname->s)
    if (H5DSset_scale(dimscaleset, nc4dimspecial) == -1) err1_break(3, "set_scale")
    if (write_dimscale_attrs(ctxid, dimscaleset, coord) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    if (space != -1) H5Sclose(space);
    if (propid != -1) H5Pclose(propid);
    if (dimscaleset != -1) H5Dclose(dimscaleset);
    if (gridgroup != -1) H5Gclose(gridgroup);
  }
  return succeeded ? 0 : -1;
}

static int create_dimscale(hid_t ctxid, const challoc *gridgrouppath, const coordinate_t *coord, hsize_t dimlength)
{
  challoc nc4dimname;
  int succeeded = 0;

  challoc_init(&nc4dimname);
  do {
    if (nc4safenamecopy(ctxid, coord->name, &nc4dimname)) err_break(3)
    if (create_dimscale2(ctxid, gridgrouppath, &nc4dimname, coord, dimlength)) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&nc4dimname);
  }
  return succeeded ? 0 : -1;
}

static int attach_dimscale(hid_t ctxid, hid_t gridgroup, const char *nc4dimname, unsigned int dimindex, hid_t dataset)
{
  hid_t dimscaleset = -1;
  int succeeded = 0;

  do {
    if ((dimscaleset = H5Dopen(gridgroup, nc4dimname, H5P_DEFAULT)) < 0) err2_break(3, "dimscale dataset open", nc4dimname)
    if (H5DSattach_scale(dataset, dimscaleset, dimindex) == -1) err2_break(3, "attach_scale", nc4dimname)

    succeeded = 1;
  } while (0);

  if (1) {
    if (dimscaleset != -1) H5Dclose(dimscaleset);
  }
  return succeeded ? 0 : -1;
}

static int create_2d_coordvar(hid_t ctxid, const challoc *gridgrouppath, const coordinate_t *coord, int ydimmajor, const dimension_t *xdim, const dimension_t *ydim, const doublealloc *data)
{
  hid_t gridgroup = -1, dimscaleset = -1, space = -1;
  challoc nc4dimname;
  int succeeded = 0;

  challoc_init(&nc4dimname);
  do {
    hid_t filetype = H5T_IEEE_F64BE;
    hid_t memtype = H5T_NATIVE_DOUBLE;
    hsize_t dimlength[2];

    if (ydimmajor) {
      dimlength[0] = ydim->dimsize;
      dimlength[1] = xdim->dimsize;
    }
    else {
      dimlength[0] = xdim->dimsize;
      dimlength[1] = ydim->dimsize;
    }

    ASSERT(data->len == dimlength[0] * dimlength[1])

    if ((gridgroup = get_h5group(ctxid, gridgrouppath->s)) == -1) err_break(3)
    if (nc4safenamecopy(ctxid, coord->name, &nc4dimname) == -1) err_break(3)

    if ((space = H5Screate_simple(2, dimlength, dimlength)) == -1) err1_break(3, "space create")
    if ((dimscaleset = H5Dcreate(gridgroup, nc4dimname.s, filetype, space, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) == -1) err1_break(3, "dimscale dataset create")
    if (H5Dwrite(dimscaleset, memtype, space, space, H5P_DEFAULT, data->s) == -1) err1_break(3, "dataset write")
    if (attach_dimscale(ctxid, gridgroup, ydim->name.s, ydimmajor ? 0 : 1, dimscaleset) == -1) err_break(3)
    if (attach_dimscale(ctxid, gridgroup, xdim->name.s, ydimmajor ? 1 : 0, dimscaleset) == -1) err_break(3)
    if (write_dimscale_attrs(ctxid, dimscaleset, coord) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    if (space != -1) H5Sclose(space);
    if (dimscaleset != -1) H5Dclose(dimscaleset);
    if (gridgroup != -1) H5Gclose(gridgroup);
    challoc_free(&nc4dimname);
  }
  return succeeded ? 0 : -1;
}

static int accumulate_grid_coordinates(hid_t ctxid, const coordinate_t *coord, challoc *attrvalue)
{
  int succeeded = 0;

  do {
    if (attrvalue->len) {
      if (!challoc_append(attrvalue, " ")) nomem1_break("coordinates")
    }
    if (!challoc_cats(attrvalue, coord->name)) nomem1_break("coordinates")

    succeeded = 1;
  } while (0);
  return succeeded ? 0 : -1;
}

static int shrink_lonlat(hid_t ctxid, const grid_t *grid, int ydimmajor, doublealloc *lons, doublealloc *lats)
{
  int succeeded = 0;

  do {
    int i;

    if (ydimmajor == 1) {
      for (i = 0; i < grid->info.ydim; ++i)
	lats->s[i] = lats->s[i * grid->info.xdim];
      lats->len = grid->info.ydim;
      lons->len = grid->info.xdim;
    }
    else if (ydimmajor == 0) {
      for (i = 0; i < grid->info.xdim; ++i)
	lons->s[i] = lons->s[i * grid->info.ydim];
      lons->len = grid->info.xdim;
      lats->len = grid->info.ydim;
    }
    else {
      ASSERT(0);
    }

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int generate_lonlat_dimension(hid_t ctxid, const challoc *gridgrouppath, const grid_t *grid, grid_aux_t *aux, challoc *coordinates)
{
  hid_t gridgroup = -1;
  dimension_t dims[2];
  doublealloc lonlat[2];
  hid_t dataset[2] = { -1, -1 };
  hid_t space[2] = { -1, -1 };
  int orthogonal = 0;
  int ydimmajor = 0;
  int succeeded = 0;

  init_dimension_t(&dims[0]);
  init_dimension_t(&dims[1]);
  doublealloc_init(&lonlat[0]);
  doublealloc_init(&lonlat[1]);
  do {
    hid_t memtype = H5T_NATIVE_DOUBLE;
    int i, failinside = 0;

    if ((gridgroup = get_h5group(ctxid, gridgrouppath->s)) == -1) err_break(3)

    if (generate_lonlat(ctxid, grid, &ydimmajor, &lonlat[0], &lonlat[1]) == -1) err_break(3)
    orthogonal = check_ortho_dimscale(grid, ydimmajor, &lonlat[0], &lonlat[1]);
    if (orthogonal) {
      if (shrink_lonlat(ctxid, grid, ydimmajor, &lonlat[0], &lonlat[1]) == -1) err_break(3)
    }

    // create two dimscales: XDim and YDim
    // if orthogonal, write partial data; otherwise, just create empty dimscale
    for (i = 0; i < 2; ++i) {
      const coordinate_t *coord;

      coord = get_coordinate(grid, orthogonal, i == 0);
      if (!challoc_copys(&dims[i].name, coord->name)) nomem1_break("dim name")
      if (nc4safename(ctxid, &dims[i].name) == -1) err_break(3)
      dims[i].dimsize = i == 0 ? grid->info.xdim : grid->info.ydim;
      if (create_dimscale2(ctxid, gridgrouppath, &dims[i].name, coord, dims[i].dimsize) == -1) err_break(3)

      if ((dataset[i] = H5Dopen(gridgroup, dims[i].name.s, H5P_DEFAULT)) == -1) err1_break(3, "dimscale dataset open")
      if ((space[i] = H5Dget_space(dataset[i])) == -1) err1_break(3, "dimscale get space")
      if (orthogonal) {
	if (H5Dwrite(dataset[i], memtype, space[i], space[i], H5P_DEFAULT, lonlat[i].s) == -1) err1_break(3, "dimscale write")
      }
    }
    if (failinside) break;

    if (!orthogonal) {
      // if dimscale is not orthogonal, full data are needed
      // due to the limit of netCDF dimscale model, these data are treated as variables rather than dimensions
      // see http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.1/cf-conventions.html
      for (i = 0; i < 2; ++i) {
	const coordinate_t *coord = get_fake_coordinate(grid, i == 0);
	if (create_2d_coordvar(ctxid, gridgrouppath, coord, ydimmajor, &dims[0], &dims[1], &lonlat[i]) == -1) err_flagbreak(3, failinside)
	if (accumulate_grid_coordinates(ctxid, coord, coordinates) == -1) err_flagbreak(3, failinside)
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (space[0] != -1) H5Sclose(space[0]);
    if (space[1] != -1) H5Sclose(space[1]);
    if (dataset[0] != -1) H5Dclose(dataset[0]);
    if (dataset[1] != -1) H5Dclose(dataset[1]);
    doublealloc_free(&lonlat[0]);
    doublealloc_free(&lonlat[1]);
    free_dimension_t(&dims[0]);
    free_dimension_t(&dims[1]);
    if (gridgroup != -1) H5Gclose(gridgroup);
  }
  if (succeeded) {
    aux->ydimmajor = ydimmajor;
    aux->orthogonal = orthogonal;
  }
  return succeeded ? 0 : -1;
}

static int convert_attr(hid_t ctxid, const char *grouppath, const char *nc4varname, const attribute_t *attr)
{
  hid_t group = -1, dataset = -1;
  int succeeded = 0;

  do {
    if ((group = get_h5group(ctxid, grouppath)) == -1) err_break(3)
    if (nc4varname) {
      if ((dataset = H5Dopen(group, nc4varname, H5P_DEFAULT)) == -1) err_break(3)
    }

    if (write_attribute(ctxid, dataset == -1 ? group : dataset, attr->name.s, attr->type, attr->value.s, attr->value.len) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    if (group != -1) H5Gclose(group);
    if (dataset != -1) H5Dclose(dataset);
  }
  return succeeded ? 0 : -1;
}

static int convert_local_attr(hid_t ctxid, const char *grouppath, const char *nc4varname, const attribute_t *attr)
{
  return convert_attr(ctxid, grouppath, nc4varname, attr);
}

static int convert_group_attr(hid_t ctxid, const char *grouppath, const attribute_t *attr)
{
  return convert_attr(ctxid, grouppath, NULL, attr);
}

static int convert_gridswath_attr(hid_t ctxid, const char *gridswathgrouppath, const attribute_t *attr)
{
  return convert_attr(ctxid, gridswathgrouppath, NULL, attr);
}

static int convert_global_attr(hid_t ctxid, const attribute_t *attr)
{
  return convert_attr(ctxid, "/", NULL, attr);
}

static int convert_attributes(hid_t ctxid, const challoc *gridswathpath, const attributealloc *gridswathattr, const challoc *groupattrpath, const attributealloc *groupattr)
{
  int succeeded = 0;

  do {
    size_t i;
    int failinside = 0;

    // grid (or swath) attribute
    for (i = 0; i < gridswathattr->len; ++i) {
      const attribute_t *attr = &gridswathattr->s[i];
      if (convert_gridswath_attr(ctxid, gridswathpath->s, attr) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    // group attribute
    for (i = 0; i < groupattr->len; ++i) {
      const attribute_t *attr = &groupattr->s[i];
      if (convert_group_attr(ctxid, groupattrpath->s, attr) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int create_var(hid_t ctxid, const challoc *grouppath, int isgrid, int isgeo, const challoc *nc4fieldname, const fieldinfo_t *field, const dimensionalloc *dimlist)
{
  hid_t group = -1, space = -1, plist = -1, dataset = -1;
  hsize_t *dimsize = NULL, *maxdimsize = NULL;
  int succeeded = 0;

  do {
    int i, failinside = 0;
    size_t j;

    if ((group = get_h5group(ctxid, grouppath->s)) == -1) err_break(3)

    if ((dimsize = malloc(sizeof(hsize_t) * field->rank)) == NULL) nomem1_break("dimsize")
    if ((maxdimsize = malloc(sizeof(hsize_t) * field->rank)) == NULL) nomem1_break("dimsize")
    for (i = 0; i < field->rank; ++i) {
      dimsize[i] = field->dims.s[i].dimsize;
      if (strcmp(field->dims.s[i].name.s, field->dims.s[i].max.s) == 0)
	maxdimsize[i] = dimsize[i];
      else {
	for (j = 0; j < dimlist->len; ++j) {
	  const dimension_t *dim = &dimlist->s[j];
	  if (strcmp(dim->name.s, field->dims.s[i].max.s) == 0) {
	    maxdimsize[i] = dim->dimsize;
	    break;
	  }
	}
	if (j == dimlist->len) err2_flagbreak(3, failinside, "cannot find max dimension", field->dims.s[i].max.s)
      }
    }
    if (failinside) break;

    if ((space = H5Screate_simple(field->rank, dimsize, maxdimsize)) == -1) err1_break(3, "space create")
    if ((plist = H5Pcreate(H5P_DATASET_CREATE)) == -1) err1_break(3, "property create")
    if (H5Pset_chunk(plist, field->rank, dimsize) == -1) err1_break(3, "chunk property")
    if (field->filler.len > 0) {
      if (H5Pset_fill_value(plist, field->type, field->filler.s) == -1) err1_break(3, "fill property")
    }
    if ((dataset = H5Dcreate(group, nc4fieldname->s, field->type, space, H5P_DEFAULT, plist, H5P_DEFAULT)) == -1) err1_break(3, "dataset create")
    {
      fieldinfo_t *mutablefield = (fieldinfo_t *)field;
      if (prepare_field_lazydata(ctxid, isgrid, isgeo, mutablefield) == -1) err2_break(3, "read data", field->name.s)
      if (H5Dwrite(dataset, field->type, space, space, H5P_DEFAULT, field->lazydata->s) == -1) err1_break(3, "dataset write")
      if (drop_field_lazydata(ctxid, mutablefield) == -1) err2_break(3, "drop data", field->name.s)
    }

    for (j = 0; j < field->localattr.len; ++j) {
      const attribute_t *attr = &field->localattr.s[j];
      if (convert_local_attr(ctxid, grouppath->s, nc4fieldname->s, attr) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (dataset != -1) H5Dclose(dataset);
    if (plist != -1) H5Pclose(plist);
    if (space != -1) H5Sclose(space);
    if (group != -1) H5Gclose(group);
    if (maxdimsize != NULL) free(maxdimsize);
    if (dimsize != NULL) free(dimsize);
  }
  return succeeded ? 0 : -1;
}

static int convert_grid_field(hid_t ctxid, const challoc *dimgrouppath, const challoc *grouppath, const fieldinfo_t *field, const grid_t *grid, int orthogonal, const challoc *coordinates)
{
  hid_t dimgroup = -1, group = -1, dataset = -1;
  challoc nc4fieldname;
  challoc nc4dimname;
  int succeeded = 0;

  challoc_init(&nc4fieldname);
  challoc_init(&nc4dimname);
  do {
    size_t i;
    int failinside = 0;

    if ((dimgroup = get_h5group(ctxid, dimgrouppath->s)) == -1) err_break(3)
    if ((group = get_h5group(ctxid, grouppath->s)) == -1) err_break(3)
    if (nc4safenamecopy(ctxid, field->name.s, &nc4fieldname) == -1) err_break(3)

    if (create_var(ctxid, grouppath, 1, 0, &nc4fieldname, field, &grid->dim) == -1) err_break(3)
    if ((dataset = H5Dopen(group, nc4fieldname.s, H5P_DEFAULT)) < 0) err2_break(3, "variable open", nc4fieldname.s)

    for (i = 0; i < field->dims.len; ++i) {
      const dimension_t *dim = &field->dims.s[i];
      const char *mangledname = dim->name.s;
      int ispredefined[2] = { 0, 0 };

      ispredefined[0] = strcmp(dim->name.s, "XDim") == 0;
      ispredefined[1] = strcmp(dim->name.s, "YDim") == 0;
      if (ispredefined[0] || ispredefined[1]) {
	const coordinate_t *coord = get_coordinate(grid, orthogonal, ispredefined[0]);
	mangledname = coord->name;
      }
      if (nc4safenamecopy(ctxid, mangledname, &nc4dimname) == -1) err_flagbreak(3, failinside)
      if (attach_dimscale(ctxid, dimgroup, nc4dimname.s, (unsigned int)i, dataset) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    if (coordinates->len) {
      if (write_attribute_string(ctxid, dataset, "coordinates", coordinates->s) == -1) err1_break(3, "coordinates attr")
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (dataset != -1) H5Dclose(dataset);
    if (group != -1) H5Gclose(group);
    if (dimgroup != -1) H5Gclose(dimgroup);
    challoc_free(&nc4fieldname);
    challoc_free(&nc4dimname);
  }
  return succeeded ? 0 : -1;
}

static int convert_swath_geofield(hid_t ctxid, const challoc *dimgrouppath, const challoc *grouppath, const fieldinfo_t *geofield, const swath_t *swath)
{
  hid_t dimgroup = -1, group = -1, dataset = -1;
  challoc nc4fieldname;
  challoc nc4dimname;
  int succeeded = 0;

  challoc_init(&nc4fieldname);
  challoc_init(&nc4dimname);
  do {
    size_t i;
    int failinside = 0;

    if ((dimgroup = get_h5group(ctxid, dimgrouppath->s)) == -1) err_break(3)
    if ((group = get_h5group(ctxid, grouppath->s)) == -1) err_break(3)
    if (nc4safenamecopy(ctxid, geofield->name.s, &nc4fieldname) == -1) err_break(3)

    if (create_var(ctxid, grouppath, 0, 1, &nc4fieldname, geofield, &swath->dim) == -1) err_break(3)
    if ((dataset = H5Dopen(group, nc4fieldname.s, H5P_DEFAULT)) < 0) err2_break(3, "variable open", nc4fieldname.s)

    for (i = 0; i < geofield->dims.len; ++i) {
      const dimension_t *dim = &geofield->dims.s[i];
      if (nc4safenamecopy(ctxid, dim->name.s, &nc4dimname) == -1) err_flagbreak(3, failinside)
      if (attach_dimscale(ctxid, dimgroup, nc4dimname.s, (unsigned int)i, dataset) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (dataset != -1) H5Dclose(dataset);
    if (group != -1) H5Gclose(group);
    if (dimgroup != -1) H5Gclose(dimgroup);
    challoc_free(&nc4fieldname);
    challoc_free(&nc4dimname);
  }
  return succeeded ? 0 : -1;
}

typedef struct {
  const fieldinfo_t *geofield;
  size_t datadimindex;
  size_t geodimindex;
  const dimmap_t *dimmap;
} swathdim_t;

void init_swathdim_t(swathdim_t *e)
{
  e->geofield = NULL;
  e->datadimindex = -1;
  e->geodimindex = -1;
  e->dimmap = NULL;
}

void free_swathdim_t(swathdim_t *e)
{
}
DECLARE_VECTOR_ALLOCATOR_FREE(swathdimalloc, swathdim_t)

static int enqueue_swathdim(hid_t ctxid, swathdimalloc *queue, const fieldinfo_t *geofield, size_t datadimindex, size_t geodimindex, const dimmap_t *dimmap)
{
  int succeeded = 0;

  do {
    swathdim_t *swathdim;
    if (!swathdimalloc_readyplus(queue, 1)) nomem1_break("swathdim_t")
    swathdim = &queue->s[queue->len++];
    init_swathdim_t(swathdim);

    swathdim->geofield = geofield;
    swathdim->datadimindex = datadimindex;
    swathdim->geodimindex = geodimindex;
    swathdim->dimmap = dimmap;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int make_augmented_dimmaps(hid_t ctxid, const swath_t *swath, const fieldinfo_t *datafield, dimmapalloc *dimmaps)
{
  int succeeded = 0;

  do {
    size_t i;
    int failinside = 0;
    dimmap_t *dimmap;

    for (i = 0; i < swath->dimmap.len; ++i) {
      const dimmap_t *sd = &swath->dimmap.s[i];
      if (!dimmapalloc_readyplus(dimmaps, 1)) nomem1_flagbreak(failinside, "dimmap")
      dimmap = &dimmaps->s[dimmaps->len++];
      init_dimmap_t(dimmap);
      if (!copy_dimmap_t(sd, dimmap)) nomem1_flagbreak(failinside, "dimmap")
    }
    if (failinside) break;

    for (i = 0; i < datafield->dims.len; ++i) {
      const dimension_t *dim = &datafield->dims.s[i];
      if (!dimmapalloc_readyplus(dimmaps, 1)) nomem1_flagbreak(failinside, "dimmap")
      dimmap = &dimmaps->s[dimmaps->len++];
      init_dimmap_t(dimmap);
      if (!challoc_copy(&dimmap->data, &dim->name)) nomem1_flagbreak(failinside, "dimmap data")
      if (!challoc_copy(&dimmap->geo, &dim->name)) nomem1_flagbreak(failinside, "dimmap geo")
      dimmap->offset = 0;
      dimmap->increment = 1;
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);
  return succeeded ? 0 : -1;
}

static int is_adjusted_dimmap(hid_t ctxid, const dimmap_t *dimmap)
{
  if (dimmap->offset == 0 && dimmap->increment == 1 && strcmp(dimmap->data.s, dimmap->geo.s) == 0) return 0;
  return 1;
}

typedef struct {
  challoc nc4dimname;
  const fieldinfo_t *geofield;
  dimmapalloc dimmaps;
} swathadjustgeo_t;

void init_swathadjustgeo_t(swathadjustgeo_t *e)
{
  challoc_init(&e->nc4dimname);
  e->geofield = NULL;
  dimmapalloc_init(&e->dimmaps);
}

void free_swathadjustgeo_t(swathadjustgeo_t *e)
{
  challoc_free(&e->nc4dimname);
  dimmapalloc_free(&e->dimmaps);
}
DECLARE_VECTOR_ALLOCATOR_FREE(swathadjustgeoalloc, swathadjustgeo_t)

static int enqueue_swathadjustgeo(hid_t ctxid, swathadjustgeoalloc *queue, const challoc *nc4dimname, const fieldinfo_t *geofield, const swathdimalloc *dimqueue, const sizetalloc *dimqueueindices)
{
  int succeeded = 0;

  do {
    size_t i;
    int failinside = 0, found;

    if (geofield->rank != dimqueueindices->len) err2_break(3, "unexpected rank", geofield->name.s)

    found = 0;
    for (i = 0; i < queue->len; ++i) {
      const swathadjustgeo_t *adjustgeo = &queue->s[i];
      if (strcmp(adjustgeo->nc4dimname.s, nc4dimname->s) == 0) {
	found = 1;
	break;
      }
    }
    if (failinside) break;

    if (!found) {
      swathadjustgeo_t *adjustgeo;
      if (!swathadjustgeoalloc_readyplus(queue, 1)) nomem1_break("swathadjustgeo_t")
      adjustgeo = &queue->s[queue->len++];
      init_swathadjustgeo_t(adjustgeo);

      if (!challoc_copy(&adjustgeo->nc4dimname, nc4dimname)) nomem1_break("adjustgeo nc4dimname")
      adjustgeo->geofield = geofield;
      if (!dimmapalloc_ready(&adjustgeo->dimmaps, geofield->rank)) nomem1_break("adjustgeo dimmaps")
      for (i = 0; i < dimqueueindices->len; ++i) {
	const dimmap_t *d = dimqueue->s[dimqueueindices->s[i]].dimmap;
	dimmap_t *dimmap = &adjustgeo->dimmaps.s[i];
	init_dimmap_t(dimmap);
	if (!copy_dimmap_t(d, dimmap)) nomem1_flagbreak(failinside, "adjustgeo dimmap")
	adjustgeo->dimmaps.len++;
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int accumulate_swath_coordinates(hid_t ctxid, const challoc *coordname, challoc *attrvalue)
{
  challoc tmp;
  int succeeded = 0;

  challoc_init(&tmp);
  do {
    int longitude = 0, latitude = 0;

    /* Since Latitude and Longitude are geo fields and they may be stored in arbitrary order, */
    /* the calling sequence of this function does not guarantee Longitude / Latitude order */
    /* For this reason, this function should be smarter than accumulate_grid_coordinates() */
    longitude = strstr(coordname->s, "Longitude") == coordname->s;
    latitude = strstr(coordname->s, "Latitude") == coordname->s;
    if (longitude) {
      if (!challoc_copy(&tmp, attrvalue)) nomem1_break("coord attr")
      if (!challoc_copys(attrvalue, coordname->s)) nomem1_break("coord attr")
      if (tmp.len) {
	if (!challoc_append(attrvalue, " ")) nomem1_break("coord attr")
	if (!challoc_cat(attrvalue, &tmp)) nomem1_break("coord attr")
      }
    }
    else if (latitude) {
      if (attrvalue->len) {
	if (!challoc_append(attrvalue, " ")) nomem1_break("coord attr")
      }
      if (!challoc_cats(attrvalue, coordname->s)) nomem1_break("coord attr")
    }

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&tmp);
  }
  return succeeded ? 0 : -1;
}

static int handle_one_cartesian_product(hid_t ctxid, const swathdimalloc *queue, const sizetalloc *queueindices, challoc *coordinates, swathadjustgeoalloc *adjustgeo)
{
  challoc geoname;
  challoc nc4dimname;
  int succeeded = 0;

  challoc_init(&geoname);
  challoc_init(&nc4dimname);
  do {
    const fieldinfo_t *geofield = queue->s[queueindices->s[0]].geofield;
    size_t i;
    int adjusted = 0, failinside = 0;

    if (!challoc_copy(&geoname, &geofield->name)) nomem1_break("mangled name")
    for (i = 0; i < queueindices->len; ++i) {
      const dimmap_t *dimmap = queue->s[queueindices->s[i]].dimmap;
      if (is_adjusted_dimmap(ctxid, dimmap))
	adjusted = 1;
      if (!challoc_catb(&geoname, "_", 1)) nomem1_flagbreak(failinside, "mangled")
      if (!challoc_catlong(&geoname, dimmap->offset, 0)) nomem1_flagbreak(failinside, "mangled")
      if (!challoc_catb(&geoname, ":", 1)) nomem1_flagbreak(failinside, "mangled")
      if (!challoc_catlong(&geoname, dimmap->increment, 0)) nomem1_flagbreak(failinside, "mangled")
    }
    if (failinside) break;

    if (adjusted) {
      if (nc4safenamecopy(ctxid, geoname.s, &nc4dimname) == -1) err_break(3)
      if (enqueue_swathadjustgeo(ctxid, adjustgeo, &nc4dimname, geofield, queue, queueindices) == -1) err_break(3)
    }
    else {
      if (nc4safenamecopy(ctxid, geofield->name.s, &nc4dimname) == -1) err_break(3)
    }

    if (accumulate_swath_coordinates(ctxid, &nc4dimname, coordinates) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&geoname);
    challoc_free(&nc4dimname);
  }
  return succeeded ? 0 : -1;
}

static int traverse_cartesian_product(hid_t ctxid, const swathdimalloc *queue, size_t geodimindex, sizetalloc *queueindices, sizetalloc *datadimindices, challoc *coordinates, swathadjustgeoalloc *adjustgeo)
{
  int succeeded = 0;

  ASSERT(geodimindex <= queueindices->len)
  ASSERT(geodimindex <= datadimindices->len)
  do {
    size_t i;
    int failinside = 0;

    for (i = 0; i < queue->len; ++i) {
      const swathdim_t *swathdim = &queue->s[i];
      const fieldinfo_t *geofield = queue->s[i].geofield;
      if (swathdim->geodimindex != geodimindex) continue;
      if (geodimindex > 0) {
	/* Only { Longitude, 0, DataTrack_lo } and { Longitude, 1, DataXTrack_lo } can be combined.
	 * If the pair was "Longitude" and "Latitude", it is not a possible combination; skip it
	 */
	const char *prevgeoname = queue->s[queueindices->s[geodimindex - 1]].geofield->name.s;
	const char *thisgeoname = geofield->name.s;
	if (strcmp(prevgeoname, thisgeoname)) continue;
	ASSERT(queue->s[queueindices->s[geodimindex - 1]].geofield == queue->s[i].geofield)
      }

      queueindices->s[geodimindex] = i;
      queueindices->len = geodimindex + 1;
      datadimindices->s[geodimindex] = swathdim->datadimindex;
      datadimindices->len = geodimindex + 1;

      /* when one concrete dimension is completed, call @handler */
      if (geodimindex == geofield->dims.len - 1) {
	if (handle_one_cartesian_product(ctxid, queue, queueindices, coordinates, adjustgeo) == -1) err_flagbreak(3, failinside)
      }
      /* go to next index */
      else if (traverse_cartesian_product(ctxid, queue, geodimindex + 1, queueindices, datadimindices, coordinates, adjustgeo) == -1) err_flagbreak(3, failinside)
    }

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int convert_swath_datafield(hid_t ctxid, const challoc *dimgrouppath, const challoc *grouppath, const fieldinfo_t *datafield, const swath_t *swath, swathadjustgeoalloc *adjustgeo)
{
  hid_t dimgroup = -1, group = -1, dataset = -1;
  challoc nc4fieldname;
  challoc nc4dimname;
  dimmapalloc augdimmaps;
  swathdimalloc queue;
  sizetalloc queueindices;
  sizetalloc datadimindices;
  challoc coordinates;
  int succeeded = 0;

  challoc_init(&nc4fieldname);
  challoc_init(&nc4dimname);
  dimmapalloc_init(&augdimmaps);
  swathdimalloc_init(&queue);
  sizetalloc_init(&queueindices);
  sizetalloc_init(&datadimindices);
  challoc_init(&coordinates);
  do {
    size_t i, j, k, l;
    int failinside = 0;

    if ((dimgroup = get_h5group(ctxid, dimgrouppath->s)) == -1) err_break(3)
    if ((group = get_h5group(ctxid, grouppath->s)) == -1) err_break(3)
    if (nc4safenamecopy(ctxid, datafield->name.s, &nc4fieldname) == -1) err_break(3)

    if (create_var(ctxid, grouppath, 0, 0, &nc4fieldname, datafield, &swath->dim) == -1) err_break(3)
    if ((dataset = H5Dopen(group, nc4fieldname.s, H5P_DEFAULT)) < 0) err2_break(3, "variable open", nc4fieldname.s)

    for (i = 0; i < datafield->dims.len; ++i) {
      const dimension_t *dim = &datafield->dims.s[i];
      if (nc4safenamecopy(ctxid, dim->name.s, &nc4dimname) == -1) err_flagbreak(3, failinside)
      if (attach_dimscale(ctxid, dimgroup, nc4dimname.s, (unsigned int)i, dataset) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    if (make_augmented_dimmaps(ctxid, swath, datafield, &augdimmaps) == -1) err_break(3)
    for (i = 0; i < datafield->dims.len; ++i) {
      const dimension_t *datafielddim = &datafield->dims.s[i];
      for (j = 0; j < augdimmaps.len; ++j) {
	const dimmap_t *dimmap = &augdimmaps.s[j];
	if (strcmp(datafielddim->name.s, dimmap->data.s)) continue;
	for (k = 0; k < swath->geofield.len; ++k) {
	  const fieldinfo_t *geofield = &swath->geofield.s[k];
	  for (l = 0; l < geofield->dims.len; ++l) {
	    const dimension_t *geofielddim = &geofield->dims.s[l];
	    if (strcmp(geofielddim->name.s, dimmap->geo.s)) continue;
	    if (enqueue_swathdim(ctxid, &queue, geofield, i, l, dimmap) == -1) err1_flagbreak(3, failinside, "enqueue")
	  }
	  if (failinside) break;
	}
	if (failinside) break;
      }
      if (failinside) break;
    }
    if (failinside) break;

    if (queue.len) {
      if (!sizetalloc_ready(&queueindices, datafield->dims.len)) err1_break(3, "queueindices")
      if (!sizetalloc_ready(&datadimindices, datafield->dims.len)) err1_break(3, "datadimindices")

      if (traverse_cartesian_product(ctxid, &queue, 0, &queueindices, &datadimindices, &coordinates, adjustgeo) == -1) err_break(3)
      if (coordinates.len > 0) {
	if (write_attribute_string(ctxid, dataset, "coordinates", coordinates.s) == -1) err1_break(3, "coordinates attr")
      }
    }

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&coordinates);
    sizetalloc_free(&queueindices);
    sizetalloc_free(&datadimindices);
    swathdimalloc_free(&queue);
    dimmapalloc_free(&augdimmaps);
    challoc_free(&nc4fieldname);
    challoc_free(&nc4dimname);
    if (dataset != -1) H5Dclose(dataset);
    if (group != -1) H5Gclose(group);
    if (dimgroup != -1) H5Gclose(dimgroup);
  }
  return succeeded ? 0 : -1;
}

static const dimension_t * lookup_dimension(const dimensionalloc *dims, const char *name)
{
  size_t i;
  for (i = 0; i < dims->len; ++i) {
    const dimension_t *dim = &dims->s[i];
    if (strcmp(dim->name.s, name) == 0)
      return dim;
  }
  return NULL;
}

typedef struct {
  long leftpoint;
  long rightpoint;
  long leftdist;
  long rightdist;
} interpolateinfo_t;

DECLARE_VECTOR_ALLOCATOR(interpolatedinfoalloc, interpolateinfo_t)

typedef struct {
  double accumulated;
} interpolatecontext_t;

static int accumulate_pivots(hid_t ctxid, const fieldinfo_t *geofield, const interpolatedinfoalloc *interpolateinfo, size_t dimindex, sizetalloc *pivotindices, interpolatecontext_t *interpolatectx)
{
  int succeeded = 0;

  do {
    int i, failinside = 0;
    size_t j;
    hid_t memtype;

    /* The returned type is what HDF-EOS5 uses to fetch data */
    if ((memtype = get_memtype(ctxid, geofield)) == -1) err_break(3)

    /* two points are used for each dimension */
    for (i = 0; i < 2; ++i) {
      int point = i == 0 ? interpolateinfo->s[dimindex].leftpoint : interpolateinfo->s[dimindex].rightpoint;
      int distsum = interpolateinfo->s[dimindex].leftdist + interpolateinfo->s[dimindex].rightdist;
      double factor;

      pivotindices->s[dimindex] = point;
      factor = i == 1 ? interpolateinfo->s[dimindex].leftdist : interpolateinfo->s[dimindex].rightdist;
      factor /= distsum;

      if (dimindex < pivotindices->len - 1) {
	/* go to next dimension index */
	interpolatecontext_t nestedctx;
	nestedctx.accumulated = 0.0f;
	if (accumulate_pivots(ctxid, geofield, interpolateinfo, dimindex + 1, pivotindices, &nestedctx) == -1) err_flagbreak(3, failinside)
	  interpolatectx->accumulated += factor * nestedctx.accumulated;
      }
      else {
	/* one of 2^n points */
	hsize_t dataindex;
	double value;

	/* treat as pivotindices->len dimensional row-major array, and calculate the index in 1-dimensional array */
	for (dataindex = j = 0; j < pivotindices->len - 1; ++j) {
	  dataindex += pivotindices->s[j];
	  dataindex *= geofield->dims.s[j + 1].dimsize;
	}
	dataindex += pivotindices->s[pivotindices->len - 1];

#ifdef EAGER_DATA_FETCH
#error This program requires lazy fetch.
#else
	/* since neither H5T_NATIVE_INT nor H5T_NATIVE_INT_g is a constant, if ... else was used */
#define HANDLE_TYPE(tid, t) else if (H5Tequal(memtype, tid) > 0) { t *v; v = (t *)&geofield->lazydata->s[dataindex * sizeof(t)]; value = *v; }
#define HANDLE_TYPE_CAST(tid, t) else if (H5Tequal(memtype, tid) > 0) { t *v; v = (t *)&geofield->lazydata->s[dataindex * sizeof(t)]; value = (double)*v; }
#endif
	if (0) { }
	HANDLE_TYPE(H5T_NATIVE_INT, int)
	HANDLE_TYPE(H5T_NATIVE_SCHAR, char)
	HANDLE_TYPE(H5T_NATIVE_SHORT, short)
	HANDLE_TYPE(H5T_NATIVE_UCHAR, char)
	HANDLE_TYPE(H5T_NATIVE_USHORT, unsigned short)
	HANDLE_TYPE(H5T_NATIVE_UINT, unsigned int)
	HANDLE_TYPE(H5T_NATIVE_LONG, long)
	HANDLE_TYPE(H5T_NATIVE_ULONG, unsigned long)
	HANDLE_TYPE_CAST(H5T_NATIVE_LLONG, long long)
	HANDLE_TYPE_CAST(H5T_NATIVE_ULLONG, unsigned long long)
	HANDLE_TYPE(H5T_NATIVE_FLOAT, float)
	HANDLE_TYPE(H5T_NATIVE_DOUBLE, double)
	HANDLE_TYPE(H5T_NATIVE_LDOUBLE, long double)
	HANDLE_TYPE(H5T_NATIVE_CHAR, char)
	else {
	  err2_flagbreak(3, failinside, "unexpected geofield type", geofield->name.s)
	}
#undef HANDLE_TYPE
	interpolatectx->accumulated += factor * value;
      }
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int finish_one_point(hid_t ctxid, const fieldinfo_t *geofield, const interpolatedinfoalloc *interpolateinfo, doublealloc *interpolated)
{
  sizetalloc pivotindices;
  int succeeded = 0;

  sizetalloc_init(&pivotindices);
  do {
    double *value;
    interpolatecontext_t interpolatectx;

    if (!doublealloc_readyplus(interpolated, 1)) nomem1_break("onepoint")
    value = &interpolated->s[interpolated->len++];

    if (!sizetalloc_ready(&pivotindices, geofield->rank)) nomem1_break("pivotindices")
    pivotindices.len = geofield->rank;
    interpolatectx.accumulated = 0.0f;
    if (accumulate_pivots(ctxid, geofield, interpolateinfo, 0, &pivotindices, &interpolatectx) == -1) err_break(3)
    *value = interpolatectx.accumulated;

    succeeded = 1;
  } while (0);

  if (1) {
    sizetalloc_free(&pivotindices);
  }
  return succeeded ? 0 : -1;
}

static long le_geoindex_from_dataindex(long dataindex, const dimmap_t *dimmap)
{
  long r = -1;
  ASSERT(dimmap->increment > 0)
  if (dimmap->offset >= 0)
    r = (dataindex - dimmap->offset) / dimmap->increment;
  else
    r = dataindex / dimmap->increment + -dimmap->offset;
  return r;
}

static long geoindex_from_dataindex(long dataindex, const dimmap_t *dimmap)
{
  long r = -1;
  ASSERT(dimmap->increment < 0)
  if (dimmap->offset >= 0)
    r = (dataindex - dimmap->offset) * -dimmap->increment;
  else
    r = dataindex * -dimmap->increment + -dimmap->offset;
  return r;
}

static long dataindex_from_geoindex(long geoindex, const dimmap_t *dimmap)
{
  long r = -1;
  ASSERT(dimmap->increment > 0)
  if (dimmap->offset >= 0)
    r = geoindex * dimmap->increment + dimmap->offset;
  else
    r = (geoindex + dimmap->offset) * dimmap->increment;
  return r;
}

static int interpolate_geofield(hid_t ctxid, const swath_t *swath, const swathadjustgeo_t *adjustgeo, size_t geodimindex, interpolatedinfoalloc *interpolateinfo, doublealloc *interpolated)
{
  int succeeded = 0;

  ASSERT_RANGE(geodimindex, 0, adjustgeo->geofield->dims.len)

  do {
    const dimension_t *geodim = &adjustgeo->geofield->dims.s[geodimindex];
    const dimmap_t *dimmap = &adjustgeo->dimmaps.s[geodimindex];
    const dimension_t *datadim = lookup_dimension(&swath->dim, dimmap->data.s);
    interpolateinfo_t *info = &interpolateinfo->s[geodimindex];
    int i, failinside = 0;

    ASSERT(strcmp(geodim->name.s, dimmap->geo.s) == 0)
    ASSERT(datadim)

    if (dimmap->increment == 0) err1_break(3, "increment is equal to 0")
    for (i = 0; i < datadim->dimsize; ++i) {
      /* if "increment" is positive, dimension of geo field has smaller number of elements */
      if (dimmap->increment > 0) {
	long geoindex = le_geoindex_from_dataindex(i, dimmap);
	/* beyond the left boundary */
	if (geoindex <= 0) {
	  info->leftpoint = 0;
	  info->rightpoint = 1;
	  if (info->rightpoint < geodim->dimsize) {
	    info->leftdist = i - dataindex_from_geoindex(info->leftpoint, dimmap);
	    info->rightdist = dataindex_from_geoindex(info->rightpoint, dimmap) - i;
	  }
	  else {
	    info->rightpoint = 0;
	    info->leftdist = -1;
	    info->rightdist = -1;
	  }
	}
	/* beyound the right boundary */
	else if (geoindex >= geodim->dimsize - 1) {
	  info->rightpoint = (long)geodim->dimsize - 1;
	  info->leftpoint = info->rightpoint - 1;
	  if (info->leftpoint >= 0) {
	    info->leftdist = i - dataindex_from_geoindex(info->leftpoint, dimmap);
	    info->rightdist = dataindex_from_geoindex(info->rightpoint, dimmap) - i;
	  }
	  else {
	    info->leftpoint = (long)geodim->dimsize - 1;
	    info->leftdist = -1;
	    info->rightdist = -1;
	  }
	}
	/* middle */
	else {
	  info->leftpoint = geoindex;
	  info->rightpoint = info->leftpoint + 1;
	  info->leftdist = i - dataindex_from_geoindex(info->leftpoint, dimmap);
	  info->rightdist = dataindex_from_geoindex(info->rightpoint, dimmap) - i;
	}
      }
      /* if "increment" is negative, dimension of data field has smaller number of elements */
      else {
	long geoindex = geoindex_from_dataindex(i, dimmap);
	/* beyond the left boundary */
	if (geoindex <= 0) {
	  info->leftpoint = 0;
	  info->rightpoint = -dimmap->increment;
	  if (info->rightpoint < geodim->dimsize) {
	    info->leftdist = geoindex - info->leftpoint;
	    info->rightdist = info->rightpoint - geoindex;
	  }
	  else {
	    info->rightpoint = 0;
	    info->leftdist = -1;
	    info->rightdist = -1;
	  }
	}
	/* beyond the right boundary */
	else if (geoindex >= geodim->dimsize - 1) {
	  info->rightpoint = (long)geodim->dimsize - 1;
	  info->leftpoint = info->rightpoint - -dimmap->increment;
	  if (info->leftpoint >= 0) {
	    info->leftdist = geoindex - info->leftpoint;
	    info->rightdist = info->rightpoint - geoindex;
	  }
	  else {
	    info->leftpoint = (long)geodim->dimsize - 1;
	    info->leftdist = -1;
	    info->rightdist = -1;
	  }
	}
	/* middle */
	else {
	  info->leftpoint = geoindex;
	  info->rightpoint = info->leftpoint + -dimmap->increment;
	  info->leftdist = 0;
	  info->rightdist = -dimmap->increment;
	}
      }

      {
	if (info->rightpoint > 0 && info->leftpoint < geodim->dimsize - 1) {
	  if (dimmap->increment > 0) {
	    ASSERT(info->leftdist + info->rightdist == dimmap->increment)
	    ASSERT(info->leftpoint + 1 == info->rightpoint)
	  }
	  else {
	    ASSERT(info->leftdist + info->rightdist == -dimmap->increment)
	    ASSERT(info->leftpoint + -dimmap->increment == info->rightpoint)
	  }
	}
	else {
	  ASSERT(info->leftpoint == info->rightpoint)
	  ASSERT(info->leftdist == -1 && info->rightdist == -1)
	}
      }

      if (geodimindex < (size_t)adjustgeo->geofield->rank - 1) {
	/* go to next dimension */
	if (interpolate_geofield(ctxid, swath, adjustgeo, geodimindex + 1, interpolateinfo, interpolated) == -1) err_flagbreak(3, failinside)
      }
      else {
	/* append one interpolated value */
	if (finish_one_point(ctxid, adjustgeo->geofield, interpolateinfo, interpolated) == -1) err_flagbreak(3, failinside)
      }
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int create_adjusted_dimscale(hid_t ctxid, const challoc *dimgrouppath, const challoc *grouppath, const swathadjustgeo_t *adjustgeo, const swath_t *swath)
{
  hid_t dimgroup = -1, group = -1, space = -1, plist = -1, dataset = -1;
  hsize_t *dimsize = NULL;
  interpolatedinfoalloc interpolateinfo;
  doublealloc interpolated;
  challoc nc4dimname;
  int succeeded = 0;

  interpolatedinfoalloc_init(&interpolateinfo);
  doublealloc_init(&interpolated);
  challoc_init(&nc4dimname);
  do {
    int i, failinside = 0;
    size_t j;
    hsize_t numelem = 1;
    hid_t memtype = H5T_NATIVE_DOUBLE;

    if ((dimgroup = get_h5group(ctxid, dimgrouppath->s)) == -1) err_break(3)
    if ((group = get_h5group(ctxid, grouppath->s)) == -1) err_break(3)

    ASSERT(adjustgeo->geofield->rank == adjustgeo->dimmaps.len)
    if ((dimsize = malloc(sizeof(hsize_t) * adjustgeo->geofield->rank)) == NULL) nomem1_break("dimsize")
    for (i = 0; i < adjustgeo->geofield->rank; ++i) {
      const dimension_t *dim = lookup_dimension(&swath->dim, adjustgeo->dimmaps.s[i].data.s);
      if (!dim) err2_flagbreak(3, failinside, "dimension does not exist", adjustgeo->dimmaps.s[i].data.s)
      dimsize[i] = dim->dimsize;
      numelem *= dimsize[i];
    }
    if (failinside) break;

    if ((space = H5Screate_simple(adjustgeo->geofield->rank, dimsize, NULL)) == -1) err1_break(3, "space create")
    if ((plist = H5Pcreate(H5P_DATASET_CREATE)) == -1) err1_break(3, "property create")
    if (H5Pset_chunk(plist, adjustgeo->geofield->rank, dimsize) == -1) err1_break(3, "chunk property")
    if ((dataset = H5Dcreate(group, adjustgeo->nc4dimname.s, adjustgeo->geofield->type, space, H5P_DEFAULT, plist, H5P_DEFAULT)) == -1) err1_break(3, "dataset create")

    if (!interpolatedinfoalloc_ready(&interpolateinfo, adjustgeo->geofield->rank)) nomem1_break("interpolate info")
    interpolateinfo.len = adjustgeo->geofield->rank;
    {
#ifndef EAGER_DATA_FETCH
      fieldinfo_t *mutablefield = (fieldinfo_t *)adjustgeo->geofield;
      if (prepare_field_lazydata(ctxid, 0, 1, mutablefield) == -1) err2_break(3, "read data", adjustgeo->geofield->name.s)
#endif
      if (interpolate_geofield(ctxid, swath, adjustgeo, 0, &interpolateinfo, &interpolated) == -1) err_break(3)
#ifndef EAGER_DATA_FETCH
      if (drop_field_lazydata(ctxid, mutablefield) == -1) err2_break(3, "drop data", adjustgeo->geofield->name.s)
#endif
    }
    ASSERT(numelem == interpolated.len)

    if (H5Dwrite(dataset, memtype, space, space, H5P_DEFAULT, interpolated.s) == -1) err1_break(3, "dataset write")

    for (j = 0; j < adjustgeo->geofield->localattr.len; ++j) {
      const attribute_t *attr = &adjustgeo->geofield->localattr.s[j];
      if (convert_local_attr(ctxid, grouppath->s, adjustgeo->nc4dimname.s, attr) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    for (i = 0; i < adjustgeo->dimmaps.len; ++i) {
      const dimmap_t *dimmap = &adjustgeo->dimmaps.s[i];
      if (nc4safenamecopy(ctxid, dimmap->data.s, &nc4dimname) == -1) err_flagbreak(3, failinside)
      if (attach_dimscale(ctxid, dimgroup, nc4dimname.s, (unsigned int)i, dataset) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (dataset != -1) H5Dclose(dataset);
    if (plist != -1) H5Pclose(plist);
    if (space != -1) H5Sclose(space);
    if (group != -1) H5Gclose(group);
    if (dimgroup != -1) H5Gclose(dimgroup);
    if (dimsize != NULL) free(dimsize);
    doublealloc_free(&interpolated);
    interpolatedinfoalloc_free(&interpolateinfo);
    challoc_free(&nc4dimname);
  }
  return succeeded ? 0 : -1;
}

static int make_groups(hid_t ctxid, const char *gridswathname, challoc *group, challoc *data, challoc *geo)
{
  hid_t groupid = -1;
  hid_t datagroupid = -1;
  hid_t geogroupid = -1;
  challoc name;
  int succeeded = 0;

  challoc_init(&name);
  do {
    if (!challoc_copys(group, gridswathname)) nomem1_break("grid path")
    if (nc4safename(ctxid, group) == -1) err_break(3)
    if ((groupid = creat_h5group(ctxid, "/", group)) == -1) err_break(3)

    if (!challoc_copys(&name, DATAGROUP_NAME)) nomem1_break("grid data path")
    if (nc4safename(ctxid, &name) == -1) err_break(3)
    if (!challoc_copy(data, group)) nomem1_break("grid data path")
    if (!challoc_cats(data, "/")) nomem1_break("grid data path")
    if (!challoc_cat(data, &name)) nomem1_break("grid data path")
    if ((datagroupid = creat_h5group(ctxid, "/", data)) == -1) err_break(3)

    if (geo) {
      if (!challoc_copys(&name, GEOGROUP_NAME)) nomem1_break("grid geo path")
      if (nc4safename(ctxid, &name) == -1) err_break(3)
      if (!challoc_copy(geo, group)) nomem1_break("grid geo path")
      if (!challoc_cats(geo, "/")) nomem1_break("grid geo path")
      if (!challoc_cat(geo, &name)) nomem1_break("grid geo path")
      if ((geogroupid = creat_h5group(ctxid, "/", geo)) == -1) err_break(3)
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (groupid != -1) H5Gclose(groupid);
    if (datagroupid != -1) H5Gclose(datagroupid);
    if (geogroupid != -1) H5Gclose(geogroupid);
    challoc_free(&name);
  }
  return succeeded ? 0 : -1;
}

int write_hdf5_onegrid(hid_t ctxid, const char *gridname, const grid_t *grid)
{
  grid_aux_t aux;
  challoc gridgrouppath;
  challoc griddatapath;
  challoc coordinates;
  int succeeded = 0;

  init_grid_aux_t(&aux);
  challoc_init(&gridgrouppath);
  challoc_init(&griddatapath);
  challoc_init(&coordinates);
  do {
    size_t i;
    int failinside = 0;

    if (make_groups(ctxid, gridname, &gridgrouppath, &griddatapath, NULL) == -1) err_break(3)

    if (generate_lonlat_dimension(ctxid, &gridgrouppath, grid, &aux, &coordinates) == -1) err_break(3)
    for (i = 0; i < grid->dim.len; ++i) {
      const dimension_t *dim = &grid->dim.s[i];
      coordinate_t coord;

      if (strcmp(dim->name.s, "XDim") == 0 || strcmp(dim->name.s, "YDim") == 0) continue;
      get_coordinate_from_dimension(dim->name.s, &coord);
      if (create_dimscale(ctxid, &gridgrouppath, &coord, dim->dimsize) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    for (i = 0; i < grid->field.len; ++i) {
      const fieldinfo_t *field = &grid->field.s[i];
      if (convert_grid_field(ctxid, &gridgrouppath, &griddatapath, field, grid, aux.orthogonal, &coordinates) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    if (convert_attributes(ctxid, &gridgrouppath, &grid->attr, &griddatapath, &grid->grpattr) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&gridgrouppath);
    challoc_free(&griddatapath);
    challoc_free(&coordinates);
    free_grid_aux_t(&aux);
  }
  return succeeded ? 0 : -1;
}

int write_hdf5_oneswath(hid_t ctxid, const char *swathname, const swath_t *swath)
{
  hid_t h5group = -1;
  challoc swathgrouppath;
  challoc swathdatapath;
  challoc swathgeopath;
  swathadjustgeoalloc adjustgeo;
  int succeeded = 0;

  challoc_init(&swathgrouppath);
  challoc_init(&swathdatapath);
  challoc_init(&swathgeopath);
  swathadjustgeoalloc_init(&adjustgeo);
  do {
    size_t i;
    int failinside = 0;

    if (make_groups(ctxid, swathname, &swathgrouppath, &swathdatapath, &swathgeopath) == -1) err_break(3)

    for (i = 0; i < swath->dim.len; ++i) {
      const dimension_t *dim = &swath->dim.s[i];
      coordinate_t coord;

      get_coordinate_from_dimension(dim->name.s, &coord);
      if (create_dimscale(ctxid, &swathgrouppath, &coord, dim->dimsize) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    for (i = 0; i < swath->geofield.len; ++i) {
      const fieldinfo_t *field = &swath->geofield.s[i];
      if (convert_swath_geofield(ctxid, &swathgrouppath, &swathdatapath, field, swath) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    for (i = 0; i < swath->datafield.len; ++i) {
      const fieldinfo_t *field = &swath->datafield.s[i];
      if (convert_swath_datafield(ctxid, &swathgrouppath, &swathdatapath, field, swath, &adjustgeo) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    for (i = 0; i < adjustgeo.len; ++i) {
      const swathadjustgeo_t *geo = &adjustgeo.s[i];
      if (create_adjusted_dimscale(ctxid, &swathgrouppath, &swathgeopath, geo, swath) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    if (convert_attributes(ctxid, &swathgrouppath, &swath->attr, &swathdatapath, &swath->grpattr) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    if (h5group != -1) H5Gclose(h5group);
    swathadjustgeoalloc_free(&adjustgeo);
    challoc_free(&swathgrouppath);
    challoc_free(&swathdatapath);
    challoc_free(&swathgeopath);
  }
  return succeeded ? 0 : -1;
}

int open_hdf5(hid_t ctxid, const char *fname)
{
  hid_t outfd = -1;
  hid_t accessprop = -1, createprop = -1;
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;

  do {
    if (ctx == NULL) err1_break(3, "invalid context id")
    if ((accessprop = H5Pcreate(H5P_FILE_ACCESS)) == -1) err1_break(3, "file access property")
#ifdef MAKE_HDFVIEW_HAPPY
    if ((outfd = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, accessprop)) == -1) err1_break(3, "H5Fcreate")
#else
//    if (H5Pset_libver_bounds(accessprop, H5F_LIBVER_LATEST, H5F_LIBVER_LATEST) == -1) err1_break(3, "lastest format property")
    if ((createprop = H5Pcreate(H5P_FILE_CREATE)) == -1) err1_break(3, "file create property")
    if (H5Pset_link_creation_order(createprop, H5P_CRT_ORDER_TRACKED | H5P_CRT_ORDER_INDEXED) == -1) err1_break(3, "link creation order")
    if (H5Pset_attr_creation_order(createprop, H5P_CRT_ORDER_TRACKED | H5P_CRT_ORDER_INDEXED) == -1) err1_break(3, "attr creation order")

    if ((outfd = H5Fcreate(fname, H5F_ACC_TRUNC, createprop, accessprop)) == -1) err1_break(3, "H5Fcreate")
#endif

    succeeded = 1;
  } while (0);

  if (1) {
    if (accessprop != -1) H5Pclose(accessprop);
    if (createprop != -1) H5Pclose(createprop);
  }
  if (succeeded) {
    ctx->outfname = fname;
    ctx->outfd = outfd;
  }
  else {
    if (outfd != -1) H5Fclose(outfd);
  }
  return succeeded ? 0 : -1;
}

int write_hdf5_globalattr(hid_t ctxid)
{
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;

  do {
    size_t i;
    int failinside = 0;

    if (ctx == NULL) err1_break(3, "invalid context id")
    for (i = 0; i < ctx->globalattr.len; ++i) {
      const attribute_t *attr = &ctx->globalattr.s[i];
      if (convert_global_attr(ctxid, attr) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);
  return succeeded ? 0 : -1;
}

void close_hdf5(hid_t ctxid)
{
  context_t *ctx = get_context(ctxid);
  if (ctx != NULL) {
    if (ctx->outfd != -1) H5Fclose(ctx->outfd);
  }
}

// vim:ts=8:sw=2:sts=2:cindent
